////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : FWSysTrayInfo
//  * Purpose   :       .
//  * Author    :  (Rouse_) 
//  * Editors   : Hint (22.02.2011), harisma(30.05.2011)
//  * Copyright :  Fangorn Wizards Lab 1998 - 2011.
//  * Version   : 1.03
//  * Home Page : http://rouse.drkb.ru
//  ****************************************************************************
//
//  Remark      :     W2k, XP  Windows 7,
//                 Windows 98     
//                  .
//
// 1.02 -     Hint.
//   Windwos Vista/Windows 7
//
// 1.03 -       ( harisma).
//        Delphi 5.
//

unit FWSysTrayInfo;

interface

uses
  Windows,
  Messages,
  SysUtils,
  Graphics,
  CommCtrl,
  Controls,
  TlHelp32,
  ShellAPI;

type
  PNprivIcon = ^TNprivIcon;
  _TNPRIVICON = record
    hWnd: THandle;
    uID: UINT;
    uCallbackMessage: UINT;
    dwState: DWORD;
    uVersion: UINT;
    hIcon: HICON;
  end;
  TNprivIcon = _TNPRIVICON;

  TFWSysIconItem = record
    uID: UINT;
    iImage: Integer;
    szTip: array [0..MAXCHAR - 1] of Char;
    dwState: DWORD;
    hWnd: THandle;
    szAppPath: string;
  end;

  TFWSysIconData = record
    Items: array of TFWSysIconItem;
    ImageList: TImageList;
  end;

  TFWSysTrayInfo = class
  private
    FSysTrayHandle, FSysTrayHandleEx: THandle;
    FIconCount, FIconCountMain: Integer;
    FIconData: TFWSysIconData;
    function GetImageList: TImageList;
    function GetIconItem(Index: Integer): TFWSysIconItem;
  protected
    function GetAppName(const Wnd: THandle): string;
    procedure DoGetSysTrayHandle; virtual;
    procedure DoGetIconCount; virtual;
    procedure DoGetIconsInfo; virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure UpdateInfo;
    procedure DeleteIco(const Index: Integer);
    property IconCount: Integer read FIconCount;
    property IconItem[Index: Integer]: TFWSysIconItem read GetIconItem;
    property ImageList: TImageList read GetImageList;
  end;

const
  {$EXTERNALSYM TBIF_BYINDEX}
  TBIF_BYINDEX            = $80000000;
  {$EXTERNALSYM TBSTATE_ENABLED}
  TBSTATE_ENABLED         = $04;
  {$EXTERNALSYM TBSTATE_HIDDEN}
  TBSTATE_HIDDEN          = $08;
  

implementation

//     :
// Shell_TrayWnd
//    |- Button                     -  
//    |- ReBarWindow32
//    |    |- MSTaskSwWClass
//    |    |    +- ToolbarWindow32  -  
//    |    +- ToolbarWindow32       -  
//    +- TrayNotifyWnd
//         |- TrayClockWClass       -     ;)
//         |- SysPager
//         |    +- ToolbarWindow32  -   
//         +- Button                -   

const
  Shell_TrayWnd   = 'Shell_TrayWnd';
  TrayNotifyWnd   = 'TrayNotifyWnd';
  SysPager        = 'SysPager';
  ToolbarWindow32 = 'ToolbarWindow32';

{ TFWSysTrayInfo }

constructor TFWSysTrayInfo.Create;
begin
  FIconCount := 0;
  FIconCountMain := 0;
  FIconData.ImageList := TImageList.Create(nil);
end;

procedure TFWSysTrayInfo.DeleteIco(const Index: Integer);
var
  IconData: TNotifyIconData;
begin
  ZeroMemory(@IconData, SizeOf(TNotifyIconData));
  IconData.cbSize := SizeOf(TNotifyIconData);
  IconData.uID := FIconData.Items[Index].uID;
  IconData.Wnd := FIconData.Items[Index].hWnd;
  Shell_NotifyIcon(NIM_DELETE, @IconData);
end;

destructor TFWSysTrayInfo.Destroy;
begin
  FIconData.ImageList.Free;
  inherited;
end;

procedure TFWSysTrayInfo.DoGetIconCount;
var
  I: Integer;
begin
  //      
  if FSysTrayHandle <> 0 then
  begin
    FIconCountMain := SendMessage(FSysTrayHandle, TB_BUTTONCOUNT, 0, 0);
    if FIconCountMain < 0 then FIconCountMain := 0;
  end
  else
    FIconCountMain := 0;

  //      ()  (Hint)
  if FSysTrayHandleEx <> 0 then
  begin
    I := SendMessage(FSysTrayHandleEx, TB_BUTTONCOUNT, 0, 0);
    if I < 0 then I := 0;
  end
  else
    I := 0;

  FIconCount := FIconCountMain + I;

  SetLength(FIconData.Items, FIconCount);
  FIconData.ImageList.Clear;
  ImageList_SetImageCount(FIconData.ImageList.Handle, FIconCount);
end;

function SetDebugPriv: Boolean;
var
  Token: THandle;
  tkp: TTokenPrivileges;
begin
  Result := false;
  if OpenProcessToken(GetCurrentProcess,
    TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY, Token) then
  begin
    if LookupPrivilegeValue(nil, PChar('SeDebugPrivilege'),
      tkp.Privileges[0].Luid) then
    begin
      tkp.PrivilegeCount := 1;
      tkp.Privileges[0].Attributes := SE_PRIVILEGE_ENABLED;
      Result := AdjustTokenPrivileges(Token, False,
        tkp, 0, PTokenPrivileges(nil)^, PDWord(nil)^);
    end;
  end;
end;

procedure TFWSysTrayInfo.DoGetIconsInfo;
var
  dwProcessID, dwBytesWriten: DWORD;
  hProcess, hImageList: THandle;
  pButtonInfo: PTBButtonInfo;
  ButtonInfo: TTBButtonInfo;
  I: Integer;
  szTextBuffer: array [0..MAXCHAR - 1] of Char;
  pszTextBuffer: Pointer;
  NprivIcon: TNprivIcon;
  SHFileInfo: TSHFileInfo;
  TmpIcon: HICON;
begin

  //    ,     
  if not SetDebugPriv then Exit;

  //  ID ,    
  GetWindowThreadProcessId(FSysTrayHandle, @dwProcessID);
  if dwProcessID = 0 then ExitProcess(GetLastError);

  //  
  hProcess := OpenProcess(PROCESS_ALL_ACCESS or PROCESS_DUP_HANDLE, TRUE, dwProcessID);
  if hProcess <> 0 then
  try
    //       
    pszTextBuffer := VirtualAllocEx(hProcess, nil, MAX_PATH,
      MEM_COMMIT, PAGE_READWRITE);
    if pszTextBuffer <> nil then
    try
      //       TTBButtonInfo
      pButtonInfo := VirtualAllocEx(hProcess, nil, SizeOf(TTBButtonInfo),
        MEM_COMMIT, PAGE_READWRITE);
      if pButtonInfo <> nil then //     GetLastError (Hint)
      try
        //  
        ZeroMemory(@ButtonInfo, SizeOf(TTBButtonInfo));
        ButtonInfo.cbSize := SizeOf(TTBButtonInfo);
        ButtonInfo.dwMask := TBIF_IMAGE or TBIF_TEXT or TBIF_STATE or TBIF_STYLE
          or TBIF_LPARAM or TBIF_COMMAND or TBIF_SIZE or TBIF_BYINDEX;
        ButtonInfo.pszText := pszTextBuffer;
        ButtonInfo.cchText := MAXCHAR;

        //      
        if not WriteProcessMemory(hProcess, pButtonInfo, @ButtonInfo,
          SizeOf(TTBButtonInfo), dwBytesWriten) then ExitProcess(GetLastError);

        //     
        for I := 0 to FIconCount - 1 do
        begin
          //       
          //     ,      (Hint)
          if I < FIconCountMain then
            SendMessage(FSysTrayHandle, TB_GETBUTTONINFO, I, Integer(pButtonInfo))
          else
            SendMessage(FSysTrayHandleEx, TB_GETBUTTONINFO, I - FIconCountMain, Integer(pButtonInfo));

          //    
          ZeroMemory(@ButtonInfo, SizeOf(TTBButtonInfo));
          ReadProcessMemory(hProcess, pButtonInfo,
            @ButtonInfo, SizeOf(TTBButtonInfo), dwBytesWriten);

          //  
          ZeroMemory(@szTextBuffer[0], MAXCHAR);
          ReadProcessMemory(hProcess, ButtonInfo.pszText,
            @FIconData.Items[I].szTip[0], ButtonInfo.cchText, dwBytesWriten);

          //     
          ZeroMemory(@NprivIcon, SizeOf(TNprivIcon));
          ReadProcessMemory(hProcess, Pointer(ButtonInfo.lParam),
            @NprivIcon, SizeOf(TNprivIcon), dwBytesWriten);

          //   
          FIconData.Items[I].uID := NprivIcon.uID;
          FIconData.Items[I].dwState := ButtonInfo.fsState;
          FIconData.Items[I].hWnd := NprivIcon.hWnd;
          FIconData.Items[I].iImage := I; //   ( 0  I-1)

          //    ,    (Hint)
          if I >= FIconCountMain then
            FIconData.Items[I].dwState := FIconData.Items[I].dwState or TBSTATE_HIDDEN;

          //      
          FIconData.Items[I].szAppPath := GetAppName(NprivIcon.hWnd);

          //   ,   -  
          //   ImageList ,   
          //    TB_GETIMAGELIST.
          //   , :
          // http://support.microsoft.com/default.aspx?scid=kb%3Ben-us%3B811415
          //     .
          //    ?
          //     ,    . (NprivIcon.hIcon)
          //         
          //   .
          //   SHGetFileInfo   SHGFI_SYSICONINDEX,
          //       ImageList
          //            
          //      ,      .
          // ..      -  
          //        ImageList
          //  :)

          ZeroMemory(@SHFileInfo, SizeOf(TSHFileInfo));
          hImageList := SHGetFileInfo(PChar(FIconData.Items[I].szAppPath),
            0, SHFileInfo, SizeOf(TSHFileInfo), SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
          ImageList_ReplaceIcon(hImageList, SHFileInfo.iIcon, NprivIcon.hIcon);
          TmpIcon := ImageList_GetIcon(hImageList, SHFileInfo.iIcon, ILD_NORMAL);
          try
            ImageList_ReplaceIcon(
              FIconData.ImageList.Handle, FIconData.Items[I].iImage, TmpIcon);
          finally
            DestroyIcon(TmpIcon);
          end;
        end;
      finally
        VirtualFreeEx(hProcess, pButtonInfo, SizeOf(TTBButtonInfo), MEM_RELEASE);
      end;
    finally
      VirtualFreeEx(hProcess, pszTextBuffer, MAX_PATH, MEM_RELEASE);
    end;
  finally
    CloseHandle(hProcess);
  end;
end;

procedure TFWSysTrayInfo.DoGetSysTrayHandle;
var
  TrayNotifyHandle, OSDependence: THandle;
begin
  FSysTrayHandle := 0; //  
  FSysTrayHandleEx := 0; //   (Windows 7)

  TrayNotifyHandle := FindWindow(Shell_TrayWnd, nil);
  if TrayNotifyHandle = 0 then Exit;
  TrayNotifyHandle := FindWindowEx(TrayNotifyHandle, 0, TrayNotifyWnd, nil);
  if TrayNotifyHandle = 0 then Exit;
  OSDependence := FindWindowEx(TrayNotifyHandle, 0, SysPager, nil);
  if OSDependence = 0 then
    //  W2K SysPager 
    OSDependence := TrayNotifyHandle;
  FSysTrayHandle := FindWindowEx(OSDependence, 0, ToolbarWindow32, nil);

  //        Windows 7,      (Hint)
  FSysTrayHandleEx := FindWindow('NotifyIconOverflowWindow', nil);
  if FSysTrayHandleEx = 0 then Exit;
  FSysTrayHandleEx := FindWindowEx(FSysTrayHandleEx, 0, ToolbarWindow32, nil);
end;

function TFWSysTrayInfo.GetAppName(const Wnd: THandle): String;
var
  hModuleSnap: THandle;
  ModuleEntry: TModuleEntry32;
  CurrentPID: DWORD;
begin
  Result := '';
  GetWindowThreadProcessId(Wnd, @CurrentPID);
  hModuleSnap := CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, CurrentPID);
  if hModuleSnap <> INVALID_HANDLE_VALUE then
  try
    ModuleEntry.dwSize := SizeOf(TModuleEntry32);
    if Module32First(hModuleSnap, ModuleEntry) then
      Result := ModuleEntry.szExePath;
  finally
    CloseHandle(hModuleSnap);
  end;
end;

function TFWSysTrayInfo.GetIconItem(Index: Integer): TFWSysIconItem;
begin
  Result := FIconData.Items[Index];
end;

function TFWSysTrayInfo.GetImageList: TImageList;
begin
  Result := FIconData.ImageList;
end;

procedure TFWSysTrayInfo.UpdateInfo;
begin
  DoGetSysTrayHandle;
  DoGetIconCount;
  DoGetIconsInfo;
end;

end.
